//  JJY Simulatro jjysim.cc
//            2010 Y.Senta

// ver 1.0:  2010.6

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include<fcntl.h>
#include<unistd.h>

#include <time.h>
#include <sys/time.h>
#include <signal.h>

#define BUFSIZE 0x1000

int g_nFD;
termios *g_pBackupTM;

char g_ComBuff[BUFSIZE];

void CloseComPort(){
  if (g_nFD){
    if (g_pBackupTM)
      tcsetattr(g_nFD, TCSANOW, g_pBackupTM);
    close(g_nFD);
    g_nFD=0;
  }
  if (g_pBackupTM){
    delete g_pBackupTM;
    g_pBackupTM=NULL;
  }
}

bool OpenComPort(const char *szDev, int bps){
  for (int i=0;i<BUFSIZE;i++)
    g_ComBuff[i]=0x00;
  
  CloseComPort();
  
  if((g_nFD = open(szDev, O_RDWR | O_NOCTTY ))==-1){
    perror(szDev);
    return false;
  }
  g_pBackupTM=new termios;
  tcgetattr(g_nFD, g_pBackupTM);
  
  termios NewTM;
  bzero(&NewTM, sizeof(NewTM));
  NewTM.c_cflag = (bps | CS8 | CLOCAL | CREAD);
  NewTM.c_oflag = 0;
  
  tcflush(g_nFD, TCIFLUSH);
  tcsetattr(g_nFD,TCSANOW,&NewTM);
  
  return true;
}

void SendComByte(int n){
  if (g_nFD)
    write(g_nFD, &g_ComBuff, n)==1;
}

//////////////////////////////////////////////////////////
#define ADJT 10
#define ONESEC 1000
#define MGNSEC 800

#define BPS 9600
#define B_MARKER (200*BPS/10/1000)
#define B_ZERO   (800*BPS/10/1000)
#define B_ONE    (500*BPS/10/1000)

#define MSEC 1  // (ONESEC/CLOCKS_PER_SEC)
#define Sleep(X) usleep(X*1000)

long   g_clock;

struct timeval g_timeval;
time_t g_initSec;

void SetDefSec(){
  gettimeofday(&g_timeval, NULL);
  g_initSec=g_timeval.tv_sec;
}

time_t Clock(){
  gettimeofday(&g_timeval, NULL);
  return (g_timeval.tv_sec-g_initSec)*1000 + g_timeval.tv_usec/1000;
}

bool WaitNextSec(){
  fflush(stdout);
  
  int c= ONESEC - ((int)Clock()*MSEC - (int)g_clock);
  if (c<0)
    return false;
  Sleep(c);
  g_clock += ONESEC;
  return true;
}

void InitTime(bool b){
  fflush(stdout);

  if (!b){
    int c=MGNSEC - ((int)Clock()*MSEC - (int)g_clock);
    if (c>0 && c <=MGNSEC)
      Sleep(c);
  }
  
  time_t t1,t2;
  
  t1=time(NULL);
  do{
    Sleep(ADJT);
    t2=time(NULL);
  } while (t1==t2);
  
  g_clock=Clock()*MSEC;
}

void signal(int){
  CloseComPort();
  fprintf(stderr, "\nExit!\n");
  _exit(0);
}

int main(int argc, char *argv[]){
  if (argc < 2 || argc >3){
    fprintf (stderr, "Usage: JJYSim COM_PORT [offset_sec]\n");
    return 0;
  }
  
  int offset=0;
  if (argc==3)
    offset=atoi(argv[2]);
  
  if (!OpenComPort(argv[1], BPS)){
    fprintf (stderr, "Can not open %s\n", argv[1]);
    return 1;
  }
  signal(SIGINT, signal);
  SetDefSec();

  int SendSig[60];
  for (int i=0;i<60;i++)
    SendSig[i]=0;
  SendSig[0]=SendSig[9]=SendSig[19]=SendSig[29]=SendSig[39]=SendSig[49]=SendSig[59]=2;

  time_t t;
  struct tm *tm;
  
  t=time(NULL)+offset;
  t++;
  tm=localtime(&t);

  fprintf(stdout, "Setting: %04d/%02d/%02d %02d:%02d:%02d\n",
	 tm->tm_year + 1900, tm->tm_mon+1, tm->tm_mday,
	 tm->tm_hour, tm->tm_min, tm->tm_sec);
  InitTime(true);
  
  for (;;) {
    t=time(NULL)+offset;
    t++;
    tm=localtime(&t);
    
    fprintf(stdout, "%02d:%02d:%02d ==> ",
	   tm->tm_hour, tm->tm_min, tm->tm_sec);
    
    if (tm->tm_sec != 0 || tm->tm_min==15 || tm->tm_min==45){
      int m=60-tm->tm_sec;
      fprintf (stdout, "Wait %02d sec...", m);
      
      for (int i=0;i<m-1;i++){
	if (!WaitNextSec()){
	  InitTime(true);
	  break;
	}
      }
      fprintf(stdout, "Done.                                        |\n");
      SendComByte(B_MARKER);
      WaitNextSec();
      
      continue;
    }
    
    int tmp;
    tmp = tm->tm_min/10;
    SendSig[1]= tmp & 4 ? 1 : 0;
    SendSig[2]= tmp & 2 ? 1 : 0;
    SendSig[3]= tmp & 1 ? 1 : 0;
    
    tmp = tm->tm_min-tmp*10;
    SendSig[5]= tmp & 8 ? 1 : 0;
    SendSig[6]= tmp & 4 ? 1 : 0;
    SendSig[7]= tmp & 2 ? 1 : 0;
    SendSig[8]= tmp & 1 ? 1 : 0;
    
    tmp = tm->tm_hour/10;
    SendSig[12]= tmp & 2 ? 1 : 0;
    SendSig[13]= tmp & 1 ? 1 : 0;

    tmp = tm->tm_hour-tmp*10;
    SendSig[15]= tmp & 8 ? 1 : 0;
    SendSig[16]= tmp & 4 ? 1 : 0;
    SendSig[17]= tmp & 2 ? 1 : 0;
    SendSig[18]= tmp & 1 ? 1 : 0;
    
    tmp = (tm->tm_yday+1)/100;
    SendSig[22]= tmp & 2 ? 1 : 0;
    SendSig[23]= tmp & 1 ? 1 : 0;
    
    tmp = (tm->tm_yday+1)/10-tmp*10;
    SendSig[25]= tmp & 8 ? 1 : 0;
    SendSig[26]= tmp & 4 ? 1 : 0;
    SendSig[27]= tmp & 2 ? 1 : 0;
    SendSig[28]= tmp & 1 ? 1 : 0;
    
    tmp = (tm->tm_yday +1)%100 - tmp*10;
    SendSig[30]= tmp & 8 ? 1 : 0;
    SendSig[31]= tmp & 4 ? 1 : 0;
    SendSig[32]= tmp & 2 ? 1 : 0;
    SendSig[33]= tmp & 1 ? 1 : 0;

    tmp=0;
    for (int i=12;i<=18;i++)
      if (SendSig[i]==1)
	tmp++;
    SendSig[36]= tmp & 1 ? 1 : 0;
    
    tmp=0;
    for (int i=1;i<=8;i++)
      if (SendSig[i]==1)
	tmp++;
    SendSig[37]= tmp & 1 ? 1 : 0;
    
    tmp = (tm->tm_year%100)/10;
    SendSig[41]= tmp & 8 ? 1 : 0;
    SendSig[42]= tmp & 4 ? 1 : 0;
    SendSig[43]= tmp & 2 ? 1 : 0;
    SendSig[44]= tmp & 1 ? 1 : 0;
    
    tmp=(tm->tm_year%100)-tmp*10;
    SendSig[45]= tmp & 8 ? 1 : 0;
    SendSig[46]= tmp & 4 ? 1 : 0;
    SendSig[47]= tmp & 2 ? 1 : 0;
    SendSig[48]= tmp & 1 ? 1 : 0;
    
    tmp=tm->tm_wday;
    SendSig[50]= tmp & 4 ? 1 : 0;
    SendSig[51]= tmp & 2 ? 1 : 0;
    SendSig[52]= tmp & 1 ? 1 : 0;
    
    InitTime(false);
    for (int i=0;i<60;i++){
      if (SendSig[i]==0){
	SendComByte(B_ZERO);
	fprintf(stdout, "_");
      } else if (SendSig[i]==1) {
	SendComByte(B_ONE);
	fprintf(stdout, "-");
      } else {
	SendComByte(B_MARKER);
	fprintf(stdout, "|");
      }
      if (i<59)
	WaitNextSec();
    }
    fprintf(stdout, "\n");
  }
  return 0;
}
